# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 14:09:07 2016

@author: Tobias Jachowski

Segments are a 2D numpy.ndarray, with a start and stop index: ([start, stop])
"""
import numpy as np
import warnings
from scipy.ndimage import maximum_filter1d
from scipy.ndimage import minimum_filter1d


def get_excited_signal(signals):
    # Determine the excited signal out of two signals
    amplitude = signals.max(axis=0) - signals.min(axis=0)
    excited_signal = amplitude.argmax()
    return excited_signal


def get_excited_sections(signals, resolution, compare_time=0.01,
                         threshold=5e-9, min_duration=0.2, validate=True):
    """
    Determine the indices (segments or range), where the first signal has a
    greater amplitude than the second signal.

    Parameters
    ----------
    signals : 2D array
        2D array of two signals to compare amplitude of.
    resolution : int
        The resolution of the `signals` in number of datapoints / s.
    compare_time : float
        Time in s for size of moving window used to compare the amplitude of
        the signals to each other.
    threshold : float
        Value, the amplitude of a signal is compared to, if only one signal is
        provided.
    min_duration : float
        Minimum duration in s a contiguous segment has to have to be finally
        accepted as one. Only used, if `validate` is True.
    validate : bool
        If True, excited segments (and non excited segments) need to be at
        least `min_duration` s.

    Returns
    -------
    tuple of segments and ranges :
    (
        (x_idx_segments, y_idx_segments),
        (x_range, y_range)
    )
    """
    window = int(np.round(compare_time * resolution))
    if window < 2:
        warnings.warn("Had to increase the compare_time for calculating "
                      "amplitudes within signal! Probably, increase the "
                      "resolution.")
        # compare_time = 2 / resolution
    window = max(2, window)

    greater = greater_amplitude(signals, window, threshold=threshold)

    min_length = 1
    if validate:
        min_length = max(1, int(np.round(min_duration * resolution)))

    x_excited = get_contiguous_segments(greater, min_length_high=min_length,
                                        min_length_low=min_length)
    if signals.ndim >= 2 and signals.shape[1] >= 2:
        y_excited = get_contiguous_segments(~greater,
                                            min_length_high=min_length,
                                            min_length_low=min_length)
    else:
        y_excited = np.empty((0, 2), dtype=int)

    x_excited_range = idx_segments_to_range(x_excited)
    y_excited_range = idx_segments_to_range(y_excited)

    return (x_excited, y_excited), (x_excited_range, y_excited_range)


def contains_greater_segment(x_segments, y_segments):
    """
    Determine, whether the first 2D segments array, contains a greater segment
    than the second 2D segments array.

    Parameters
    ----------
    x_segments : 2D segments array
        2D array with two columns. The first column contains the start and the
        second column the corresponding stop indices.
    y_segments : 2D segments array
        2D array with two columns. The first column contains the start and the
        second column the corresponding stop indices.

    Returns
    -------
    bool
        True, if x_segments contains a greater segment than y_segments.
        Otherwise, False.

    Examples
    --------
    positions = view.get_data(traces='positionXY', copy=False)
    x, y = sn.get_excited_sections(positions, view.samplingrate)[0]
    x_is_excited = sn.contains_greater_segment(x, y)
    """
    max_x = 0
    for x in x_segments:
        max_x = max(max_x, x[1] - x[0])
    max_y = 0
    for y in y_segments:
        max_y = max(max_y, y[1] - y[0])
    return max_x > max_y


def greatest_amplitude(signals, size, threshold=5e-9):
    """
    Return 1D array with indices of the signal with the greatest amplitude.
    If signals.ndim == 1 or signals.shape[1] == 1 returns an array of boolean
    values which are True, if amplitude of signal is > threshold.

    Parameters
    ----------
    signals : 2D array
        2D array of a signal for which the amplitudes should be compared. If
        only one signal is given, the amplitude is compared to threshold.
        Otherwise, the amplitude is compared to the other amplitudes of the
        other signals.
        First dimension datapoints, second dimension different signals.
    size : int
        Size of the window the amplitudes should be calculated from and
        compared to each other. Needs to be >= 2.
    threshold : float
        Value, the amplitude of a signal is compared to, if only one signal is
        provided.

    Returns
    -------
    1D numpy.ndarray dtype=int or dtype=bool
        Index of where amplitude of first signal (first axis) is greater than
        the other signals or True where amplitude is greater than the
        threshold.
    """
    if size <= 1:
        raise ValueError("Size should be at least 2, to calculate amplitudes "
                         "within the signals!")
    amplitude = (maximum_filter1d(signals, size, axis=0)
                 - minimum_filter1d(signals, size, axis=0))
    if signals.ndim == 1 or signals.shape[1] == 1:
        greatest = amplitude > threshold
        greatest.shape = -1
    else:
        greatest = np.argmax(amplitude, axis=1)

    return greatest


def greater_amplitude(signals, size, threshold=5e-9):
    """
    Determine, where a signal has a greater amplitude than a given threshold,
    or alternatively, if given, other signals.

    Parameters
    ----------
    signals : 2D array
        2D array of a signal for which the amplitudes should be compared. If
        only one signal is given, the amplitude is compared to `threshold`.
        Otherwise, the amplitude is compared to the other amplitudes of the
        other signals.
        First dimension datapoints, second dimension different signals.
    size : int
        Size of the window the amplitudes should be calculated from and
        compared to each other. Needs to be >= 2.
    threshold : float
        Value, the amplitude of a signal is compared to, if only one signal is
        provided.

    Returns
    -------
    1D numpy.ndarray dtype=bool
        True where amplitude of first signal (first axis) is greater than the
        other signals or the threshold.
    """
    greater = greatest_amplitude(signals, size, threshold=threshold)
    # Convert the index arrays into boolean arrays, True, where first signal
    # has greatest ampitude
    if signals.ndim >= 2 and signals.shape[1] >= 2:
        greater = (greater == 0)
    return greater


def get_contiguous_segments(plateaus, min_distance_center=1, min_length_high=1,
                            min_length_low=1, check_center_distance_first=True,
                            fuse=True, validate_truncated=True):
    # original source derived from
    # http://stackoverflow.com/questions/4494404/find-large-number-of-consecutive-values-fulfilling-bool_array-in-a-numpy-array
    # modified and extended by Tobias Jachowski
    """
    Find contiguous True values (plateau) of the boolean array `plateaus`.
    Return a 2D array where the first column is the start index of the region
    and the second column is the stop index (segments).

    Parameters
    ----------
    plateaus : np.ndarray of type bool
    min_distance_center : int
        Minimum distance, the center of two successive plateaus need to be
        apart from each other. If the distance is shorter, the stop of the
        leading and the start of the following plateau are neglected,
        effectively fusing the two plateaus.
    min_length_high : int
        Minimum number of contiguous True values a plateau has to have to get
        detected as one. If the plateau ist shorter, it is assumed to be no
        plateau and is therefore neglected. The minimum length of the plateaus
        usually is checked, after the distance of the center was checked.
        See parameter `check_center_distance_first`.
    min_length_low : int
        Minimum number of contiguous False values a valley has to have to get
        detected as one. If the valley is shorter, it is assumed to be part of
        a greater plateau, consisting of the preceding plateau, the short
        valley itself, and the succeeding plateau and is therefore neglected.
        The length of valleys is checked after the minimum length of the
        plateaus was checked.
    check_center_distance_first : bool
        Check `min_distance_center` before `min_length_high`.
    fuse : bool
        Fuse plateaus with a too short center distance, or delete the ones
        following the other ones too close.
    validate_truncated : bool, optional
        Validate the length of plateaus starting or stopping at the very first
        [0] or last [-1] index of the plateaus array, i.e. are possibly
        truncated plateaus. If set to True, even these plateaus need to have at
        least `min_length_high` datapoints. If set to False, even truncated -
        and therefore possibly too short - plateaus are detected.

    Returns
    -------
    2D numpy.ndarray of type int
        Segments of plateaus with start indices in the first and stop indices
        in the second column.
    """

    if plateaus.size == 0:
        return np.empty((0, 2), dtype=int)

    # Find the indices of changes in "plateaus"
    d = np.diff(plateaus)
    idx, = d.nonzero()

    # We need to start things after the change in "plateaus". Therefore,
    # we'll shift the index by 1 to the right.
    idx += 1

    lock_first = False
    if plateaus[0]:
        # If the start of plateaus is True prepend a 0
        idx = np.r_[0, idx]
        lock_first = not validate_truncated

    lock_last = False
    if plateaus[-1]:
        # If the end of plateaus is True, append the length of the array
        idx = np.r_[idx, plateaus.size]
        lock_last = not validate_truncated

    # No contiguous segments detected
    if idx.size < 2:
        return np.empty((0, 2), dtype=int)

    idx_start = idx[:-1:2]
    idx_stop = idx[1::2]

    def check_center_distance(idx_start, idx_stop):
        if min_distance_center == 1:
            return idx_start, idx_stop

        # Create a linked list from the two arrays describing the start and
        # stop inidices of the plateaus. This implementation avoids the (for
        # long lists) expensive pop() function of lists.
        class Plateau(object):
            def __init__(self, start, stop, next=None):
                self.next = next
                self._start = start
                self._stop = stop
                self.center = start + (stop - start) / 2
            @property
            def start(self):
                return self._start
            @start.setter
            def start(self, start):
                self._start = start
                self.center = self.start + (self.stop - self.start) / 2
            @property
            def stop(self):
                return self._stop
            @stop.setter
            def stop(self, stop):
                self._stop = stop
                self.center = self.start + (self.stop - self.start) / 2
        current = Plateau(idx_start[-1], idx_stop[-1])
        for start, stop in zip(idx_start[-2::-1], idx_stop[-2::-1]):
            current = Plateau(start, stop, current)
        first = current

        # Iteratively check distance of the center of one plateau (start, stop)
        # to the following one, and either fuse them or delete the next one, if
        # the distance is too small
        num_plateaus = len(idx_start)
        while current.next is not None:
            if current.next.center - current.center < min_distance_center:
                if fuse:
                    # Correct the stop of the leading plateau to be the one
                    # of the following and implicitly correct the center
                    # position of the now bigger leading plateau
                    current.stop = current.next.stop
                # Delete the following plateau and its corresponding center
                current.next = current.next.next
                # Correct number for deleted plateau
                num_plateaus -= 1
            else:
                # select the next plateau
                current = current.next

        # Create numpy arrays from linked list
        start = np.empty(num_plateaus)
        stop = np.empty(num_plateaus)
        current = first
        i = 0
        while current is not None:
            start[i] = current.start
            stop[i] = current.stop
            current = current.next
            i += 1

        return start, stop

    def check_length_high(idx_start, idx_stop):
        if min_length_high == 1:
            return idx_start, idx_stop
        # Check length of plateaus first
        # (stop_high - start_high) or (uneven_index - even_index) >= min_length
        plateau = idx_stop - idx_start >= min_length_high
        # If stops are locked, make sure to reset them to True
        plateau[0] = plateau[0] or lock_first
        plateau[-1] = plateau[-1] or lock_last
        # Select only accepted starts/stops (plateaus)
        idx_start = idx_start[plateau]
        idx_stop = idx_stop[plateau]
        return idx_start, idx_stop

    checks = [check_length_high, check_center_distance]
    if check_center_distance_first:
        checks.reverse()
    for check in checks:
        idx_start, idx_stop = check(idx_start, idx_stop)

    # Only one plateau detected, ignore all valleys
    if idx_start.size <= 1:
        idx = np.sort(np.r_[idx_start, idx_stop])
        idx.shape = (-1, 2)
        return idx

    # Check length of valleys
    # (stop_low - start_low) or (even_index - uneven_index) >= min_length
    valley = np.where(idx_start[1:] - idx_stop[:-1] >= min_length_low)[0]
    # Select only accepted stops/starts (valleys) and keep first start and last
    # stop of plateaus
    start = idx_start[np.r_[0, valley + 1]]
    stop = idx_stop[np.r_[valley, idx_stop.size - 1]]

    # Concatenate and sort passed start indices
    idx = np.sort(np.r_[start, stop])
    # Reshape the result into two columns
    idx.shape = (-1, 2)

    return idx


def get_steps(signal, resolution, step_time=0.1, min_dwell_time=0.0):
    """
    Detect positive and negative steps in a signal.

    Parameters
    ----------
    signal : 1D numpy.ndarray
        Signal that contains steps to be detected.
    resolution : float
        Resolution of the signal in Hz.
    step_time : float
        The time a step takes in s. Should at least be greater as the greatest
        period that can be expected due to noise.
    min_dwell_time : float
        Min time between two succesive steps in s. If dwell time between two
        pre_detected steps is shorter the second step will be ignored.

    Returns
    -------
    tuple of 1D int index arrays
        (pos_steps, neg_steps)
    """
    window = max(1, int(step_time * resolution))
    min_dwell_length = max(1, int(min_dwell_time * resolution))
    return _get_steps(signal, window, min_dwell_length)


def _get_steps(signal, window, min_dwell_length=1):
    X = signal
    Xl = X[:-window]
    Xr = X[window:]

    pos_step_plateaus = Xl < Xr
    neg_step_plateaus = Xl > Xr

    # The detected plateaus caused by the steps should have a length of at
    # least size window (with a factor of safety due to noise)
    min_length_high = window * 0.95
    # The detected valleys caused by the dwell time (dwell length) are
    # shortened by a value of size window.
    min_length_low = max(1, (min_dwell_length - window) * 0.95)

    pos_steps = get_contiguous_segments(pos_step_plateaus,
                                        min_length_high=min_length_high,
                                        min_length_low=min_length_low,
                                        validate_truncated=True)
    neg_steps = get_contiguous_segments(neg_step_plateaus,
                                        min_length_high=min_length_high,
                                        min_length_low=min_length_low,
                                        validate_truncated=True)

    # The stop indices is the start of every new dwell plateau (directly
    # after the step)
    # pos_steps = pos_steps[:,1]
    # neg_steps = neg_steps[:,1]

    # The mean of start and stop is window/2 left of the middle of the step
    pos_steps = np.round(pos_steps.mean(axis=1) + window/2).astype(int)
    neg_steps = np.round(neg_steps.mean(axis=1) + window/2).astype(int)

    return pos_steps, neg_steps


def get_extrema(signal, resolution=1, highest_frequency=1,
                reduce_false_positives=False, idx_plateaus=False):
    """
    Determine the indices of Minima und Maxima of a signal.

    Parameters
    ----------
    signal : 1D numpy.ndarray
        The signal to be analysed.
    resolution : float
        Resolution of the signal (e.g. in Hz).
    highest_frequency : float
        Highest expected frequency in Hz of occurence of minima/maxima in the
        signals. If you set a too low frequency, extrema will not be detected.
        A frequency to high will result in to many falsly positive detected
        extrema. See also parameter `reduce_false_extrema`.
    reduce_false_positives : bool
        Set to True, to reduce falsly positive detected extrema. As a side
        effect, the extrema within half a wavelength of the given highest
        frequency at the ends of the signal could be shifted or even not be
        detected. Make sure to properly set `highest_frequency`, first.
        Another option to reduce False positives is to smoothen the signal,
        before using this function.
    idx_plateaus : bool
        If idx_plateaus is True, return not only the indices of the extrema,
        but also the start and the stop indices (segments) of the plateaus
        corresponding to the detected extrema.

    Returns
    -------
    tuple of numpy.ndarray of int
        Indices of detected minima and maxima (minima, maxima).
    """
    # The time should be adjusted according to the expected frequency and
    # should be < 1/2 of the period of the highest frequency expected (one half
    # wave equals to one extremum, factor=0.5). This time is used to calculate
    # a window size, for which the values of two datapoints are compared to be
    # greater or smaller -> extrema detection.
    # For reliable extremum detection at the ends of the signal, time should
    # even be less (factor=0.05).
    # Time <- depending on noise and amplitude, too!
    # -> higher frequency/amplitude, smaller time
    # greater time -> less false positives
    if reduce_false_positives:
        factor = 0.45  # < 50% of the period (95% of 50%)
    else:
        factor = 0.05  # <= 5% of the period
    time = factor/highest_frequency
    window = max(1, int(time * resolution))
    return _get_extrema(signal, window, idx_plateaus=idx_plateaus)


def _get_extrema(signal, window, idx_plateaus=False):
    """
    Get attributes of a signal like pattern X:

    Parameters
    ----------
    signal
    window : int
        should be greater than 50 points for A = 6 µm and f = 0.2 Hz, because
        the noise of the piezo feed back is such that it takes roughly 50
        datapoints to clearly discriminate between a data point with a true
        local maximum value and a data point with a local maximum due to noise
        (plateau of ~50) Should be smaller than datapoints for one half period,
        too. 100 is a good value for triangle signal 400 is a save value for
        sinusiodal signal 350 works for sinusiodal signal
    idx_plateaus : bool
        If idx_plateaus is True, return not only the indices of the extrema,
        but also the start and the stop indices (segments) of the plateaus
        corresponding to the detected extrema.

    Returns
    -------
    tuple : tuple of numpy.ndarrays of type int
        Tuple of arrays of minima and maxima.
    """
    # For future implementations of _get_extrema() or another implementation,
    # probably consider one of the solutions presented in
    # https://blog.ytotech.com/2015/11/01/findpeaks-in-python/
    X = signal

    Xl = X[:-2*window]
    Xm = X[window:-window]
    Xr = X[2*window:]

    # find indices of data points, which are smaller/greater than the
    # left/right ones:
    # left datapoint = comparepoint - window
    # right datapoint = comparepoint + window
    # This will create a plateau:
    minima_plateaus = np.logical_and(Xm < Xr, Xm < Xl)
    maxima_plateaus = np.logical_and(Xm > Xr, Xm > Xl)

    # Get start/stop indices of plateaus, make sure they have at least a length
    # of window and are at least a length of window apart, accept truncated
    # extrema plateaus, if they start or end at the very first [0] or last [-1]
    # index of the plateaus array.
    # Consider noise to shorten the length of plateaus and valleys to up to 5 %
    min_length_high = window * 0.95
    min_length_low = window * 0.95

    minima = get_contiguous_segments(minima_plateaus,
                                     min_length_high=min_length_high,
                                     min_length_low=min_length_low,
                                     validate_truncated=False)
    maxima = get_contiguous_segments(maxima_plateaus,
                                     min_length_high=min_length_high,
                                     min_length_low=min_length_low,
                                     validate_truncated=False)

    # correct the indices for the shift due to truncating Xm (Xl and Xr)
    minima += window
    maxima += window

    # calculate center position of start and stop indices
    minima_c = np.round(minima.mean(axis=1)).astype(int)
    maxima_c = np.round(maxima.mean(axis=1)).astype(int)

    if idx_plateaus:
        return minima_c, maxima_c, minima, maxima

    return minima_c, maxima_c


def get_value_segments(signal, value=0.0, std_noise=0.01, max_length=17,
                       min_length=1):
    """
    Search for the locations where the signal is equal to value and return
    index segments of these locations.

    Parameters
    ----------
    value : float, optional
        The value that should be looked up in `signal`.
    std_noise : float, optional
        Root mean square of the noise of the signal.
    max_length : int, optional
        Max number of contiguous datapoints that are allowed to be outside the
        `value` +/- `std_noise` range, but still be assumed to be equal to the
        `value`. On average, one out of 1 / (0.32^`max_length`) datapoints will
        falsly be detected as not beeing equal to the `value`. Assuming the
        `std_noise` was set correctly, this would be 1 out of ~ 258,000,000
        datapoints for the default value 17.
    min_length : int, optional
        Min number of contiguous datapoints that need to be within the range
        `value` +/- `std_noise`, to be detected as equal to `value`.

    """
    # Get the boolean idx array of the datapoints beeing within 0 +/-
    # std_noise.
    s = signal - value
    base_g = s > - std_noise
    base_s = s < std_noise
    base = np.logical_and(base_g, base_s)
    base = base.flatten()
    seg = get_contiguous_segments(base, min_length_low=max_length,
                                  min_length_high=min_length)
    return seg


def get_segment_centers(segments):
    if len(segments) == 0:
        return np.empty((0, 2))
    diff = segments[:, 1] - segments[:, 0]
    centers = np.round((segments[:, 0] + diff / 2)).astype(int)
    return centers


def center_sampled_segments(segments, centers=None, decimate=1, as_list=False,
                            slices=False):
    if as_list:
        idx = []
    else:
        if slices:
            dtype = object
        else:
            dtype = int
        idx = np.empty(0, dtype=dtype)

    if slices:
        f_idx = slice
    else:
        f_idx = np.arange

    if centers is None:
        centers = get_segment_centers(segments)

    idx = [f_idx(segment[0] + (center - segment[0]) % decimate or decimate,
                 segment[1],
                 decimate)
           for segment, center in zip(segments, centers)]

    if not as_list and not slices:
        idx = np.hstack(idx)

    return idx


def baseline_idx(signals, value=0.0, std_noise=0.01, max_length=17,
                 decimate=1):
    """
    Calculate the indices of where the `signals` are equal to `value` within
    a given standard deviation `std_noise`.
    This method is based on the function `signal.get_value_segments()` and
    `signal.center_sampled_segments()`.

    Parameters
    ----------
    signals : 2D numpy.ndarray of type float
        Array containing the signals in the columns.
    value : float, optional
        See function `pyoti.evaluate.signal.get_value_segments()`
    std_noise : float, optional
        See function `pyoti.evaluate.signal.get_value_segments()`
    max_length : int, optional
        See function `pyoti.evaluate.signal.get_value_segments()`
    decimate : int, optional
        See function `pyoti.evaluate.signal.center_sampled_segments()`
    """
    if signals.ndim == 1:
        signals = signals[:, np.newaxis]
    idx = np.empty(0, dtype=int)
    for signal in signals.T:
        segments = get_value_segments(signal, value=value, std_noise=std_noise,
                                      max_length=max_length)
        centers = get_segment_centers(segments)
        i = center_sampled_segments(segments, centers=centers,
                                    decimate=decimate)
        idx = np.r_[idx, i]
    idx.sort()
    return idx


def get_sections(signal, minima, maxima):
    """
    Separate a signal into rising (rise) and falling (fall) sections.
    And in turn separate these sections into positive (pos) and
    negative (neg) sections.

    Returns
    -------
    tuple of segments and ranges :
    (
        (posrise, posfall, negfall, negrise, pos, neg, rise, fall),
        (posrise_range, posfall_range, negfall_range, negrise_range,
         pos_range, neg_range, rise_range, fall_range)
    )
    """
    X = signal
    pos = X > 0
    neg = ~ pos

    # Get the indices of the change of sign.
    # The indices should have a difference of period/2 (except the difference
    # from first and to last datapoint).
    # A minimum length of period/3 should be save to exclude small local
    # sign changes, but still detect true (global) sign changes
    # Include the very first datapoint of pos/neg range and shift index by 1
    # min_period = get_period(minima, maxima, mode='min')
    # length = int(np.round(min_period/3))
    pos = get_contiguous_segments(pos)
    neg = get_contiguous_segments(neg)

    # Make sure the ranges are contiguous and not interrupted like the original
    # ones received by comparison of sign of X.
    # pos_range = np.hstack([ np.arange(change_of_sign[i],
    #                                   change_of_sign[i + 1])
    #                         for i in np.arange(0,
    #                                            change_of_sign.size - 1,
    #                                            2)])
    # neg_range = np.hstack([ np.arange(change_of_sign[i],
    #                                   change_of_sign[i + 1])
    #                         for i in np.arange(1,
    #                                            change_of_sign.size - 1,
    #                                            2)])
    pos_range = idx_segments_to_range(pos)
    neg_range = idx_segments_to_range(neg)

    # calculate all extrema, extend first and last extremum and avoid invalid
    # indices
    extrema = np.sort(np.r_[minima, maxima])

    if extrema.size == 0:
        rise_range = np.empty((0, 2), dtype=int)
        fall_range = np.empty((0, 2), dtype=int)
        rise = np.empty((0, 2), dtype=int)
        fall = np.empty((0, 2), dtype=int)
    else:
        if extrema.size == 1:
            # Only one maximum or minimum -> no info about period.
            # Take all indices before and after extremum
            first = 0
            last = X.size
        else:
            max_period = get_period(minima, maxima, mode='max')
            half_period = int(np.round(max_period / 2))
            first = max(0, extrema[0] - half_period)
            last = min(X.size, extrema[-1] + half_period)

        extrema = np.r_[first, extrema, last]

        # Assume very fist extremum is (appended) minimum:
        # maxima[0] < minima[0]
        # -> extremum[0] -> extremum[1] = rise
        if extrema.size % 2 == 0:
            idx_rise = slice(0, None)
            idx_fall = slice(1, -1)
        else:
            idx_rise = slice(0, -1)
            idx_fall = slice(1, None)

        rise = extrema[idx_rise].reshape(-1, 2)
        fall = extrema[idx_fall].reshape(-1, 2)

        # Very first extremum is (appended) maximum
        # exchange fall and rise
        if ((maxima.size == 1 and extrema.size == 1) or
           (maxima.size >= 1 and minima.size >= 1 and minima[0] < maxima[0])):
            rise, fall = fall, rise

        rise_range = idx_segments_to_range(rise)
        fall_range = idx_segments_to_range(fall)

    posrise_range = np.intersect1d(rise_range, pos_range)
    posfall_range = np.intersect1d(fall_range, pos_range)
    negfall_range = np.intersect1d(fall_range, neg_range)
    negrise_range = np.intersect1d(rise_range, neg_range)

    posrise = idx_range_to_segments(posrise_range)
    posfall = idx_range_to_segments(posfall_range)
    negfall = idx_range_to_segments(negfall_range)
    negrise = idx_range_to_segments(negrise_range)

    # posrise = []
    # for stop in np.intersect1d(rise, pos_range + 1):
    #     idx = int(np.floor(np.where(np.in1d(rise, stop))[0] / 2))
    #     start = pos_range[
    #                 np.min(np.intersect1d(np.where(pos_range <= stop)[0],
    #                              np.where(pos_range >= rise[idx, 0])[0]))]
    #     #if start != stop:
    #     posrise.append(np.array([start, stop]))
    # posrise = np.array(posrise)
    # posfall = []
    # for start in np.intersect1d(fall, pos_range):
    #     idx = int(np.floor(np.where(np.in1d(fall, start))[0] / 2))
    #     stop = pos_range[
    #                np.max(np.intersect1d(np.where(pos_range >= start)[0],
    #                             np.where(pos_range <= fall[idx, 1])[0]))] + 1
    #     #if start != stop:
    #     posfall.append(np.array([start, stop]))
    # posfall = np.array(posfall)
    # negfall = []
    # for stop in np.intersect1d(fall, neg_range + 1):
    #     idx = int(np.floor(np.where(np.in1d(fall, stop))[0] / 2))
    #     start = neg_range[
    #                 np.min(np.intersect1d(np.where(neg_range <= stop)[0],
    #                              np.where(neg_range >= fall[idx, 0])[0]))]
    #     #if start != stop:
    #     negfall.append(np.array([start, stop]))
    # negfall = np.array(negfall)
    # negrise = []
    # for start in np.intersect1d(rise, neg_range):
    #     idx = int(np.floor(np.where(np.in1d(rise, start))[0] / 2))
    #     stop = neg_range[
    #                np.max(np.intersect1d(np.where(neg_range >= start)[0],
    #                             np.where(neg_range <= rise[idx, 1])[0]))] + 1
    #     #if start != stop:
    #     negrise.append(np.array([start, stop]))
    # negrise = np.array(negrise)

    return ((posrise, posfall, negfall, negrise, pos, neg, rise, fall),
            (posrise_range, posfall_range, negfall_range, negrise_range,
             pos_range, neg_range, rise_range, fall_range)
            )


def get_period(minima, maxima, mode='mean'):
    """
    Calculate the period of given indices of extrema (minima and maxima) of an
    oscillation.

    Parameters
    ----------
    minima : 1D array
        1D array of indices of minima of an oscillation.
    maxima : 1D array
        1D array of indices of minima of an oscillation.
    mode : str
        Return the 'mean', 'max', 'min', or 'median' period.

    Returns
    -------
    float or int
        length of the period of the oscillation
    """
    extrema = np.sort(np.r_[minima, maxima])
    periods = np.diff(extrema) * 2

    if periods.size >= 1:  # at least 2 extrema
        if mode == 'mean' or mode == 'average':
            period = np.mean(periods)
        elif mode == 'max':
            period = np.max(periods)
        elif mode == 'min':
            period = np.min(periods)
        else:  # 'median'
            period = np.median(periods)
    else:
        period = 0

    return period


def idx_to_idx_segments(idx, start=None, stop=None):
    """
    Create stop indices
    Concatenate and sort start (and stop) indices
    """
    if idx.size >= 2:
        idx_segs = np.sort(np.r_[idx, idx[1:-1]])
    else:  # size is 1 or empty
        idx_segs = np.array([], dtype=int)

    if idx.size >= 2:
        if start is not None and start < idx[0]:
            idx_segs = np.r_[start, idx[0], idx_segs]
        if stop is not None and stop > idx[-1]:
            idx_segs = np.r_[idx_segs, idx[-1], stop]
    if idx.size == 1:
        if start is not None and start < idx[0]:
            idx_segs = np.r_[start, idx[0], idx_segs]
        if stop is not None and stop > idx[-1]:
            idx_segs = np.r_[idx_segs, idx[-1], stop]

    # Reshape the result into two columns
    idx_segs.shape = (-1, 2)
    return idx_segs


def idx_segments_to_range(segments, decimate=None):
    """
    Convert a 2D segments array into a 1D range array.
    """
    idx = np.empty(0, dtype=int)
    if segments.size > 0:
        idx = np.hstack([np.arange(segment[0], segment[1], decimate)
                        for segment in segments])
    return idx


def idx_segments_to_slices(segments, decimate=None):
    """
    Convert a 2D segments array into a 1D array of slices.
    """
    if decimate is not None:
        decimate = int(np.round(decimate))
    slices = []
    if segments.size > 0:
        slices = np.array([slice(segment[0], segment[1], decimate)
                          for segment in segments])
    return slices


def idx_range_to_segments(range):
    """
    Convert a 1D range array into a 2D segments array.
    """
    segments = np.empty((0, 2), dtype=int)
    if range.size >= 1:
        idx = np.where(np.diff(range) > 1)[0]
        start = np.r_[range[0], range[idx + 1]]
        stop = np.r_[range[idx], range[-1]] + 1
        segments = np.sort(np.r_[start, stop])
        segments.shape = (-1, 2)
    return segments


def get_min_size_segments(segments, size):
    segments = segments[segments[:, 1] - segments[:, 0] >= size]
    return segments


def shift_and_size_segments(segments, shift, size, min_start=-1, max_stop=0):
    segments = shift_segments(segments, shift)
    segments = size_segments(segments, size)

    segments = limit_segments(segments, min_start=min_start, max_stop=max_stop)

    return segments


def size_segments(segments, size, min_start=-1, max_stop=0):
    if size > 0:
        segments = np.sort(np.r_[segments[:, 0], segments[:, 0] + size])
        segments.shape = (-1, 2)

    segments = limit_segments(segments, min_start=min_start, max_stop=max_stop)

    return segments


def shift_segments(segments, shift, min_start=-1, max_stop=0):
    # shift ("rotate") the indices of segments
    segments = segments + shift

    segments = limit_segments(segments, min_start=min_start, max_stop=max_stop)

    return segments


def limit_segments(segments, min_start=-1, max_stop=0):
    # create copy to protect outer array
    segments = segments.copy()

    if segments.ndim == 1:
        # if no segments but indices
        if min_start > -1:
            segments[segments < min_start] = min_start
        if max_stop > 0:
            segments[segments > max_stop - 1] = max_stop - 1
    else:
        if min_start > -1:
            starts = segments[:, 0]
            stops = segments[:, 1]
            starts[starts < min_start] = min_start
            stops[stops < min_start + 1] = min_start + 1
        if max_stop > 0:
            starts = segments[:, 0]
            stops = segments[:, 1]
            starts[starts > max_stop - 1] = max_stop - 1
            stops[stops > max_stop] = max_stop

    return segments


def get_excited_offset(signal, minima, maxima, period):
    X = signal

    # Calculate voltage (offset) of signal
    offset_min_pos = minima + np.round(period/4)
    left_first_min_offset = minima[0] - np.round(period/4)
    if left_first_min_offset > 0:
        offset_min_pos = np.append(left_first_min_offset, offset_min_pos)
    offset_min_pos = offset_min_pos[offset_min_pos < X.shape[0]].astype(int)

    offset_max_pos = maxima + np.round(period/4)
    left_first_max_offset = maxima[0] - np.round(period/4)
    if left_first_max_offset > 0:
        offset_max_pos = np.append(left_first_max_offset, offset_max_pos)
    offset_max_pos = offset_max_pos[offset_max_pos < X.shape[0]].astype(int)

    offset = np.mean(np.append(X[offset_min_pos], X[offset_max_pos]))

    return offset


def outliers(signal, iqr_factor=1.5):
    # Filter out outliers
    # Points with mean intensities that are further than 1.5 times the IQR
    # away from 1st or 3rd quartile
    # iqr_factor = iqr_factor # standard 1.5, strong 3.0
    qrt1 = np.nanpercentile(signal, 25, interpolation='midpoint')
    qrt3 = np.nanpercentile(signal, 75, interpolation='midpoint')
    iqr = qrt3 - qrt1
    upper = qrt3 + iqr * iqr_factor
    lower = qrt1 - iqr * iqr_factor
    wo_outls = np.logical_and(signal <= upper, signal >= lower)
    w_outls = np.logical_or(signal > upper, signal < lower)
    return wo_outls, w_outls


def validate_extrema(extrema):
    # Rewrite this faulty function!!!
    return
    # ### exclude outliers
    # exclude outliers by calculating the quartiles
    # if the signal has a constant frequency the distance of one extremum to
    # the next one should be constant
    diff = extrema[1:] - extrema[:-1]
    median = np.median(diff)
    lower = 0.95 * median
    upper = 1.05 * median
    idx, = np.where(np.logical_and(diff > lower, diff < upper))
    if idx.size == 0:
        idx = slice(0, 1, 1)
    else:
        idx = np.append(idx, idx[-1] + 1)
    return extrema[idx]



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 15:31:12 2016

@author: Tobias Jachowski
"""
import numpy as np

from pyoti.calibration.calibsource import CalibrationSource


class CNMatlabSource(CalibrationSource):
    def __init__(self, filename=None, **kwargs):
        if filename is None:
            raise TypeError("CNMatlabSource missing the required positional "
                            "argument 'filename'.")
        self.filename = filename

        calib = np.genfromtxt(self.filename, skip_header=6,
                              invalid_raise=False)
        self.corrfactor = calib[0]
        self.dsurf = calib[1] * 1e-6  # µm -> m
        self.beta = calib[2:5] * 1e-9  # nm/V -> m/V
        self.kappa = calib[5:8] * 1e-3  # pN/nm -> N/m
        self.mbeta = calib[8:11] * 1e-3  # * 1e-9 * 1e6  # nm/V/µm -> m/V/m
        self.mkappa = calib[11:14] * 1e3  # * 1e-3 * 1e6 pN/nm/µm -> N/m/m
        self.radiusspec = calib[14] * 1e-6  # µm -> m
        self.focalshift = calib[15]
        self.name = "MATLAB calibration file originally loaded from \n    %s" \
                    % (self.filename)


class CNParaSource(CalibrationSource):
    def __init__(self, filename=None, beta=None, kappa=None, name=None,
                 **kwargs):
        """
        beta: displacement sensitivity in nm/mV
        kappa: stiffness in pN/nm
        """
        if filename is None:
            raise TypeError("CNParaSource missing the required positional "
                            "argument 'filename'.")
        self.filename = filename
        para = np.loadtxt(self.filename, comments='%', delimiter='\t')
        beta = beta or para[3:6] * 1e-9  # nm/V -> m/V
        kappa = kappa or para[6:9] * 1e-3  # pN/nm -> N/m
        name = name or 'Cellular Nanoscience parameter file originally ' \
                       'loaded from \n    %s' % (self.filename)
        super().__init__(beta=beta, kappa=kappa, name=name, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep  7 18:13:42 2017

@author: tobiasj
"""

import importlib
import sys
import os
import appdirs


pyoti_plugins_dirs = [os.path.dirname(globals()['__file__'])]
site_plugins_dirs = [os.path.join(directory, 'pyoti') for directory in
                         appdirs.AppDirs('pyoti',
                                    multipath=True).site_data_dir.split(':')]
user_plugins_dirs = [os.path.join(directory, 'plugins') for directory in
                         appdirs.AppDirs('pyoti',
                                    multipath=True).user_data_dir.split(':')]
plugins_kinds = ['datasources', 'modifications', 'calibsources']


def print_dirs():
    for kind in plugins_kinds:
        for dirs in [user_plugins_dirs, site_plugins_dirs, pyoti_plugins_dirs]:
            for d in dirs:
                print(os.path.join(d, kind))


def load_module(module_path, module_name):
    # Load module from source file [1], [2], and [3].
    # Set the __name__ attribute of the module and the __module__ attribute of
    # all classes to 'module_name'.
    # The names of the classes of objects, consisting of their class and module
    # names, are stored in the ZODB database, permanently [4]!
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Map old modules to new module names that have already been loaded [5].
    # Objects are unpickled by looking up the classes of the corresponding
    # module.
    sys.modules[module_name] = module

    # [1] https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    # [2] https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path#67692
    # [3] https://stackoverflow.com/questions/19009932/import-arbitrary-python-source-file-python-3-3#19011259
    # [4] http://www.zodb.org/en/latest/tutorial.html#introduction
    # [5] https://docs.python.org/3/library/sys.html


def load_modules():
    for plug_dir in pyoti_plugins_dirs + site_plugins_dirs + user_plugins_dirs:
        for kind in plugins_kinds:
            path = os.path.join(plug_dir, kind)
            if os.path.isdir(path):
                for name in os.listdir(path):
                    if name[-3:] != '.py' or name == '__init__.py':
                        continue
                    mod_path = os.path.join(plug_dir, kind, name)
                    mod_name = ''.join(['pyoti.plugins.',
                                        kind, '.',
                                        name[:-3]])
                    load_module(mod_path, mod_name)


# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 21:36:44 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np

from pyoti.modification.modification import Modification, GraphicalMod
from pyoti import helpers as hp
from pyoti.evaluate import signal as sn
from pyoti.evaluate import tether as tr


class IRotation(GraphicalMod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Define some attributes needed for the graphical visualization of the
        # spline

        # plateaus [excited_psd, excited_position] for left and right stress
        # regions
        self.rightextension = None
        self.leftextension = None
        self.rightforce = {}
        self.leftforce = {}

        # Dimensions of the buttons to adjust the plateaus:
        self.left = 0.1
        self.bottom = 0.79
        self.width = 0.0625
        self.height = 0.046875
        self.bspace = 0.01
        self.lspace = 0.01333

        # create some properties of actions/corresponding buttons
        action  = ['upX',   'downX', 'upY',   'downY', 'upZ',   'downZ', 'rotlX', 'rotrX', 'rotlY', 'rotrY', 'rotlZ', 'rotrZ']
        label   = ['upX',   'downX', 'upY',   'downY', 'upZ',   'downZ', 'rotlX', 'rotrX', 'rotlY', 'rotrY', 'rotlZ', 'rotrZ']
        offsetX = [-0.0025,  0.0025,  0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.0   ]
        offsetY = [ 0.0,     0.0,    -0.0025,  0.0025,  0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.0   ]
        offsetZ = [ 0.0,     0.0,     0.0,     0.0,    -0.0025,  0.0025,  0.0,     0.0,     0.0,     0.0,     0.0,     0.0   ]
        rotateX = [ 0.0,     0.0,     0.0,     0.0,     0.0,     0.0,     0.5,    -0.5,     0.0,     0.0,     0.0,     0.0   ]
        rotateY = [ 0.0,     0.0,     0.0,     0.0,     0.0,    -0.0,     0.0,     0.0,     0.5,    -0.5,     0.0,     0.0   ]
        rotateZ = [ 0.0,     0.0,     0.0,     0.0,     0.0,    -0.0,     0.0,     0.0,     0.0,     0.0,     0.5,    -0.5   ]
        row     = [ 1,       0,      -6,      -7,      -6,      -7,       1,       0,       1,       0,       1,       0     ]
        column  = [ 6,       6,       0,       0,       6,       6,       0,       0,       1,       1,       2,       2     ]
        self.action  = action
        self.label   = dict(list(zip(action, label)))
        self.offsetX = dict(list(zip(action, offsetX)))
        self.offsetY = dict(list(zip(action, offsetY)))
        self.offsetZ = dict(list(zip(action, offsetZ)))
        self.rotateX = dict(list(zip(action, rotateX)))
        self.rotateY = dict(list(zip(action, rotateY)))
        self.rotateZ = dict(list(zip(action, rotateZ)))
        self.row     = dict(list(zip(action, row)))
        self.column  = dict(list(zip(action, column)))

        self._axes = ['3D', 'X', 'Y', 'Z']

        self._lines = {}
        self.supertitle = None
        self._button = {}

    def _figure(self):
        """
        Interactive determination of rotation angle.
        """
        # Adjust the angles for rotation of of the QPD signal

        # create new figure and axes for adjusting the angles
        plot_params = {'wspace': 0.09, 'hspace': 0.08}
        self._set_plot_params(plot_params)
        figure, ax_ = plt.subplots(2, 2, sharex=True, sharey=False)

        ax = dict(list(zip(self._axes, ax_.flatten())))

        # Create buttons for interactive correction of plateaus and assign
        # correct functions
        # See http://math.andrej.com/2009/04/09/pythons-lambda-is-broken/ for
        # explanation of weird function assignment
        ax_button = {}
        for ac in self.action:
            ax_button[ac] \
                = figure.add_axes([self.column[ac] *
                                   (self.lspace + self.width) +
                                   self.left, self.row[ac] *
                                   (self.bspace + self.height) +
                                   self.bottom,
                                   self.width,
                                   self.height])
            self._button[ac] = plt.Button(ax_button[ac], self.label[ac])

            def ap(event, ac=ac):
                self._adjust_plateaus(ac)
            # connect button to action, accordingly
            self._button[ac].on_clicked(ap)

        # create lines to plot the data
        for plot in self._axes:
            self._lines['left' + plot] = ax[plot].plot([0], [0], 'r',
                                                       alpha=0.75)[0]
            self._lines['right' + plot] = ax[plot].plot([0], [0], 'g',
                                                        alpha=0.75)[0]
            ax[plot].ticklabel_format(useOffset=False)
            ax[plot].grid(True)
        ax['3D'].set_ylabel('Force (pN)')
        ax['Y'].set_ylabel('Force (pN)')
        ax['Y'].set_xlabel('Extension (nm)')
        ax['Z'].set_xlabel('Extension (nm)')

        self.supertitle = figure.suptitle("Adjust plateaus to make them "
                                          "overlap")

        return figure

    def _update_fig(self, **kwargs):
        """
        Update the plot
        """
        # prepare data for interactive plot
        self.calculate_plateaus()

        for plot in self._axes:
            l_line = self._lines['left' + plot]
            r_line = self._lines['right' + plot]

            l_line.set_data(self.leftextension * 1e9,
                            self.leftforce[plot] * 1e12)
            r_line.set_data(self.rightextension * 1e9,
                            self.rightforce[plot] * 1e12)
            # recompute ax.dataLim
            l_line.axes.relim()
            # update ax.viewLim using new dataLim
            l_line.axes.autoscale_view()

    def _pre_close_fig(self):
        # Store attachment fit plot for ducumentation
        self.supertitle.set_text('Adjusted  plateaus')
        self._lines.clear()
        self._button.clear()
        self.supertitle = None

    def _adjust_plateaus(self, action):
        """
        Adjusts the attachment (offset of excited_position) and the the scaling
        factor to correct for differences of left and right DNA overstretching
        plateaus. It is interactively called from the data plot (see below) and
        updates the plot accordingly.
        """
        # change offset and scaling for plateaus
        self.modification.iattributes.offsetPsdX += self.offsetX[action]
        self.modification.iattributes.offsetPsdY += self.offsetY[action]
        self.modification.iattributes.offsetPsdZ += self.offsetZ[action]

        self.modification.iattributes.angleX += self.rotateX[action]
        self.modification.iattributes.angleY += self.rotateY[action]
        self.modification.iattributes.angleZ += self.rotateZ[action]

        # recalculate the plateaus with the new offset and scaling values
        self.update_fig()

    def calculate_plateaus(self):
        """
        Calculate the plateaus according to the offsets and the angles of
        data_based.
        """
        # recalculate data for plotting
        traces = ['psdX', 'psdY', 'psdZ', 'positionX', 'positionY',
                  'positionZ']
        data_based = self.modification._get_data_based(
            traces=traces, window=False, decimate=True)

        # subtract offsets
        data_based[:, 0] -= self.modification.iattributes.offsetPsdX
        data_based[:, 1] -= self.modification.iattributes.offsetPsdY
        data_based[:, 2] -= self.modification.iattributes.offsetPsdZ

        # set positionZ and calibration, needed by
        # self.rotate -> self.rot_factor
        # make sure that both position signal and calibration are taken form
        # self.view_apply
        positionZ = data_based[:, hp.slicify(5)]
        calibration = self.modification.view_based.calibration

        # rotate psd
        data_based[:, 0], data_based[:, 1], data_based[:, 2] \
            = self.modification.rotate(
                data_based[:, 0], data_based[:, 1],
                data_based[:, 2], positionZ, calibration)

        # set some variable names for easy access of data_based
        psdXYZ = data_based[:, hp.slicify([0, 1, 2])]
        positionXYZ = data_based[:, hp.slicify([3, 4, 5])]
        positionXY = data_based[:, hp.slicify([3, 4])]
        positionZ = data_based[:, hp.slicify(5)]

        # calculate extension
        distanceXYZ = tr.distanceXYZ(calibration, psdXYZ, positionXYZ)
        distance = tr.distance(distanceXYZ, positionXY)
        extension = tr.extension(distance, calibration.radius)

        # calculate force
        displacementXYZ = calibration.displacement(psdXYZ, positionZ=positionZ)
        # Get the force acting in the same direction as the displacement
        fXYZ = calibration.force(displacementXYZ, positionZ=positionZ)

        force = tr.force(tr.forceXYZ(calibration, psdXYZ, positionZ),
                         positionXY)
        force = {'3D': force,
                 'X': fXYZ[:, 0],
                 'Y': fXYZ[:, 1],
                 'Z': fXYZ[:, 2]}

        # determine regions where DNA is stretched to the right and left side
        ex = self.modification._excited()
        excited_position = self.modification._NAME['position'][ex]
        positionEl = traces.index(excited_position)
        signal = data_based[:, positionEl]  # [positionE]
        resolution = self.modification.view_based.samplingrate \
            / self.modification.decimate
        minima, maxima = sn.get_extrema(signal, resolution)
        rightstress, _, leftstress \
            = sn.get_sections(signal, minima, maxima)[1][0:3]

        # set plateau data arrays
        self.rightextension = extension[rightstress]
        self.leftextension = extension[leftstress]
        for plot in self._axes:
            self.rightforce[plot] = force[plot][rightstress]
            self.leftforce[plot] = force[plot][leftstress]


class Rotation(Modification):
    GRAPHICALMOD = IRotation

    def __init__(self, **kwargs):
        traces_apply = ['psdX', 'psdY', 'psdZ']
        super().__init__(datapoints=12000, traces_apply=traces_apply, **kwargs)

        # Define parameters that are used to calculate the modification
        # the angles the ellipsoid has to be rotated
        # rotation around X
        self.add_iattribute('angleX', description='Angle about X (deg)',
                            value=0.0)
        self.add_iattribute('angleY', description='Angle about Y (deg)',
                            value=0.0)
        self.add_iattribute('angleZ', description='Angle about Z (deg)',
                            value=0.0)
        # offset of PSD relative to trap center position of bead
        self.add_iattribute('offsetPsdX', description='Offset PSD X (V)',
                            value=0.0)
        self.add_iattribute('offsetPsdY', description='Offset PSD Y (V)',
                            value=0.0)
        self.add_iattribute('offsetPsdZ', description='Offset PSD Z (V)',
                            value=0.0)

        # Parameters for rotation matrix calculation
        self.rotation_method = 'm'  # 'N', 'm' or 'V'

    def _print_info(self):
        print(("    Rotation is in '%s' space" % self.rotation_method))

    def _modify(self, data, samples, data_traces, data_index, mod_index):

        # Get data that is needed for the rotation modification, but is not
        # contained in the data array, that is requested to be getting modified
        # (`data_traces` in `data`)
        needed_traces = self.traces_apply  # traces_apply
        needed_traces.append('positionZ')  # psdX, psdY, psdZ, positionZ
        # calculate missing traces
        extra_traces = hp.missing_elements(needed_traces, data_traces)
        extra_data = self._get_data_apply(samples=samples, traces=extra_traces,
                                          copy=False)

        # combine assigned data and traces with extra fetched data and traces
        traces_tuple = (data_traces, extra_traces)
        data_tuple = (data, extra_data)

        # function to easily get data for a trace from different data
        def get_target_data(target_trace, traces_tuple, data_tuple):
            target_data = None
            for i, traces in enumerate(traces_tuple):
                if target_trace in traces:
                    index = traces.index(target_trace)
                    target_data = data_tuple[i][:, index]
            return target_data

        # get psdX, psdY and psdZ
        l_data = {}
        for trace in self.traces_apply:
            l_data[trace] = get_target_data(trace, traces_tuple, data_tuple)

        # correct for offset of psds
        x = self.iattributes.offsetPsdX
        y = self.iattributes.offsetPsdY
        z = self.iattributes.offsetPsdZ
        if x != 0.0:
            l_data['psdX'] = l_data['psdX'] - x
        if y != 0.0:
            l_data['psdY'] = l_data['psdY'] - y
        if z != 0.0:
            l_data['psdZ'] = l_data['psdZ'] - z

        # positionZ and calibration, needed by
        # self.rotate -> self.rot_factor
        # make sure that both position signal and calibration are taken form
        # self.view_apply
        positionZ = get_target_data('positionZ', traces_tuple, data_tuple)
        calibration = self.view_apply.calibration

        # rotate the data
        l_data['psdX'], l_data['psdY'], l_data['psdZ'] \
            = self.rotate(l_data['psdX'], l_data['psdY'], l_data['psdZ'],
                          positionZ, calibration)

        for trace in self.traces_apply:
            if trace in data_traces:
                index = data_traces.index(trace)
                data[:, index] = l_data[trace]

        return data

    def rotate(self, data_x, data_y, data_z, positionZ, calibration):
        ax = self.iattributes.angleX
        ay = self.iattributes.angleY
        az = self.iattributes.angleZ

        if ax == 0.0 and ay == 0.0 and az == 0.0:
            return (data_x, data_y, data_z)

        x = ax * np.pi / 180.0
        y = ay * np.pi / 180.0
        z = az * np.pi / 180.0

        cf = self.calibration_factor(positionZ, calibration)
        rf = self.rot_factor

        # https://www.siggraph.org/education/materials/HyperGraph/modeling/mod_tran/3drota.htm
        # angle about axis ...
        x_o = data_x
        y_o = data_y
        z_o = data_z

        # right handed coordinate system
        # rotation is counterclockwise about the axis coming out of the image
        # plane
        # z should always be <= 0 (negative), because the bead is stressed down
        #
        # rotate about x
        #
        #    z |
        #      |
        #      |
        #      |_______ y
        #      /
        #     /
        #  x /
        #
        if x == 0.0:
            y_2 = y_o
            z_2 = z_o
        else:
            cos_x = np.cos(x)
            sin_x = np.sin(x)
            y_2 = y_o * cos_x - z_o * rf(2, 1, cf) * sin_x
            z_2 = z_o * cos_x + y_o * rf(1, 2, cf) * sin_x

        # rotate about y
        #
        #    x |
        #      |
        #      |
        #      |_______ z
        #      /
        #     /
        #  y /
        #
        if y == 0.0:
            z_o = z_2
            x_2 = x_o
        else:
            cos_y = np.cos(y)
            sin_y = np.sin(y)
            z_o = z_2 * cos_y - x_o * rf(0, 2, cf) * sin_y
            x_2 = x_o * cos_y + z_2 * rf(2, 0, cf) * sin_y

        # rotate about z
        #
        #    y |
        #      |
        #      |
        #      |_______ x
        #      /
        #     /
        #  z /
        #
        if z == 0.0:
            x_o = x_2
            y_o = y_2
        else:
            cos_z = np.cos(z)
            sin_z = np.sin(z)
            x_o = x_2 * cos_z - y_2 * rf(1, 0, cf) * sin_z
            y_o = y_2 * cos_z + x_2 * rf(0, 1, cf) * sin_z

        return (x_o, y_o, z_o)

    def calibration_factor(self, positionZ, calibration):
        # position signal, to calculate height dependent calibration factors
        if positionZ.ndim < 2:
            positionZ = positionZ[:, np.newaxis]

        # rotation in m space
        if self.rotation_method == 'm':
            # beta in m/V, displacement sensitivity
            beta = calibration.beta(positionZ)[:, 0:3]
            factor = beta
        # rotation in pN space
        elif self.rotation_method == 'N':
            # beta in m/V, displacement sensitivity
            beta = calibration.beta(positionZ)[:, 0:3]
            # N/m, stiffness
            kappa = calibration.kappa(positionZ)[:, 0:3]
            factor = kappa * beta
        else:
            # self.rotation_method == 'V'  # rotation in V space
            factor = 1.0

        return factor

    def rot_factor(self, a, b, calibration_factor):
        """
        Calculates a weighting factor for rotating the 3D QPD signal in V via
        the signal in N (stiffness * displacement_sensitivity), m
        (displacement_sensitivity), or V space.
        (If you rotate a vector in N space, the absolute value (force) should
        stay the same, regardless of the axis. A rotation in V space would lead
        to bias, due to the different displacement sensitivities and
        stiffnesses of all three axes.)
        To calculate the factors, this method uses the calibration of the
        view, this modification is applied to.
        """
        cf = calibration_factor
        if self.rotation_method == 'm' or self.rotation_method == 'N':
            factor = cf[:, a] / cf[:, b]
        else:
            factor = 1.0

        return factor

    def Rxyz(self, positionZ, calibration, x=0.0, y=0.0, z=0.0):
        """
        Create rotationmatrix along x, y and z in degrees.
        Rotation for data with N samples and 3 dimensions (XYZ):
            R = Rxyz(x,y,z)
            data_rot = np.dot(data, R.T)

        Rxyz takes the mean() of the rotation factors, calculated by
        self.rot_factor()!
        """

        # angle about axis ...
        x = x * np.pi / 180.0
        y = y * np.pi / 180.0
        z = z * np.pi / 180.0

        f = self.rot_factor  # calculate weighting factors

        fXY = f(0, 1, positionZ, calibration).mean()
        fXZ = f(0, 2, positionZ, calibration).mean()
        fYX = f(1, 0, positionZ, calibration).mean()
        fYZ = f(1, 2, positionZ, calibration).mean()
        fZX = f(2, 0, positionZ, calibration).mean()
        fZY = f(2, 1, positionZ, calibration).mean()

        Rx = np.matrix([[     1.0,            0.0,            0.0       ],
                        [     0.0,            np.cos(x), -fZY*np.sin(x) ],
                        [     0.0,        fYZ*np.sin(x),      np.cos(x) ]])
        Ry = np.matrix([[     np.cos(y),      0.0,       -fZX*np.sin(y) ],
                        [     0.0,            1.0,            0.0       ],
                        [ fXZ*np.sin(y),      0.0,            np.cos(y) ]])
        Rz = np.matrix([[     np.cos(z), -fYX*np.sin(z),      0.0       ],
                        [ fXY*np.sin(z),      np.cos(z),      0.0       ],
                        [     0.0,            0.0,            1.0       ]])

        return Rx*Ry*Rz


# The following is only to update to database version 0.8.0
class GRotation(Rotation):
    pass



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 14 11:50:29 2014

@author: Tobias Jachowski
"""
__author__ = "Tobias Jachowski"
__copyright__ = "Copyright 2016, The PyOTIC Project"
__credits__ = []
__license__ = "Apache-2.0"
__maintainer__ = "Tobias Jachowski"
__email__ = "pyoti@jachowski.de"
__status__ = "beta"


import os
directory = os.path.dirname(globals()['__file__'])

try:
    with open(os.path.join(directory, 'VERSION.txt')) as f:
        __version__ = f.read().strip()
except:
    with open(os.path.join(directory, '..', 'VERSION.txt')) as f:
        __version__ = f.read().strip()


def version():
    """
    Returns the version.
    """
    return __version__


import sys

if 'IPython' in sys.modules:
    # We are in an ipython/juypter environment
    ip = get_ipython()
    if 'ZMQInteractiveShell' in repr(ip):
        # We are in a notebook or jupyter lab,
        # try to use ipympl backend (module://ipympl.backend_nbagg)
        try:
            import matplotlib
            matplotlib.use('module://ipympl.backend_nbagg')
        except ImportError:
            pass
            # default to inline in kernel environments
            # if hasattr(ip, 'kernel'):
            #     print('enabling inline matplotlib')
            #     ip.enable_matplotlib('inline')
            # else:
            #     print('enabling matplotlib')
            #     ip.enable_matplotlib()

    # Set format for inline plots
    from IPython.display import set_matplotlib_formats
    # %config InlineBackend.figure_formats = ['png']
    set_matplotlib_formats('png', 'svg', 'pdf', 'jpeg', quality=90)

# Load pyoti plugins
from .plugins import plugin_loader
plugin_loader.load_modules()


from . import experiment as ep
from . import evaluate as ev


def info():
    print("PyOTI - the investigator package of the PyOTIC software")
    print("Version: %s" % version())
    print()
    nb_path = os.path.abspath(".")
    print("The actual working path is: '%s'" % nb_path)


def create_experiment(**kwargs):
    """
    Create a new Experiment or get the last created one.
    Optionally, open an experiment file `filename'.
    For documentation see experiment.create_experiment()
    """
    return ep.create_experiment(**kwargs)


def open_experiment(filename=None, directory=None, **kwargs):
    """
    Create a new Experiment or get the last created one.
    Optionally, open an experiment file `filename'.
    For documentation see experiment.open_experiment()
    """
    return ep.open_experiment(filename=filename, directory=directory, **kwargs)


def close_experiment():
    """
    Close the last created Experiment.
    For documentation see experiment.close_experiment()
    """
    return ep.close_experiment()


def save_experiment():
    """
    Save the last created Experiment.
    For documentation see experiment.save_experiment()
    """
    return ep.save_experiment()


def create_calibration(**kwargs):
    """
    Create a Calibration object.
    For documentation see experiment.create_calibration()
    """
    return ep.create_calibration(**kwargs)


def create_tether(**kwargs):
    """
    Create a Tether object.
    """
    return ev.Tether(**kwargs)


def create_motion(**kwargs):
    """
    Create a Motion object.
    """
    return ev.Motion(**kwargs)


def create_stepped(**kwargs):
    raise DeprecationWarning("The function `create_stepped()` was replaced "
                             "by the function `create_motion()`.")



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 14 23:31:01 2016

@author: Tobias Jachowski
"""

import configparser
import importlib
import os
import appdirs


pyoti_config_dirs = [os.path.join(os.path.dirname(globals()['__file__']),
                                  'configs')]
site_config_dirs = appdirs.AppDirs('pyoti',
                                   multipath=True).site_config_dir.split(':')
user_config_dirs = appdirs.AppDirs('pyoti',
                                   multipath=True).user_config_dir.split(':')


def print_dirs():
    for dirs in [user_config_dirs, site_config_dirs, pyoti_config_dirs]:
        for d in dirs:
            print(d)


def _cfg_file_paths(cfgfile):
    cfgfiles = []
    for directory in pyoti_config_dirs + site_config_dirs + user_config_dirs:
        file_path = os.path.join(directory, cfgfile)
        if file_path not in cfgfiles:
            cfgfiles.append(file_path)
    return cfgfiles


def read_cfg_file(cfgfile, verbose=True):
    """
    Read in a configuration file and return a dictionary.

    Parameters
    ----------
    cfgfile : str
        The path to the configfile.
    verbose : bool, optional
        Print message, if config file was not found.

    Returns
    -------
    configparser.ConfigParser()
        Includes all sections (keys1) and options (keys2) from the configfile:
        { key1: { key2 : value } }
    """
    cfg = configparser.ConfigParser()
    cfg.optionxform = str
    cfgfiles = _cfg_file_paths(cfgfile)
    cfgfiles_found = cfg.read(cfgfiles)
    if len(cfgfiles_found) == 0:
        raise FileNotFoundError('No such file or directory: %s' % cfgfiles)
    return cfg


def get_cfg_option(cfg, sec, opt, verbose=False):
    """
    Retrieve value of a specific option of a configuration.

    Parameters
    ----------
    cfg : configparser.ConfigParser()
        Configuration as retrieved by the function read_cfg_file().
    sec : str
        The section in which the option is located.
    opt : str
        The option that should be retrieved.
    verbose : bool, optional
        Print info, if either section or option could not be found in cfg.

    Returns
    -------
    str
        Value of the option
    """
    if sec not in cfg:
        verbose and print("Section '%s' is not in configuration '%s'"
                          % (sec, cfg))
        return None

    if opt not in cfg[sec]:
        verbose and print("Option '%s' is not in section '%s'"
                          % (opt, sec))
        return None

    option = cfg[sec][opt]

    return option


def get_cfg_sec_dict(cfg, sec, convert='float', verbose=False):
    """
    Retrieve a dictionary of a section with options as keys and corresponding
    values.

    Parameters
    ----------
    cfg : configparser.ConfigParser()
        Configuration as retrieved by the function read_cfg_file().
    sec : str
        The section in which the options are located.
    convert : str
        The type of the values of the options.
    verbose : bool, optional
        no function so far ...

    Returns
    -------
    dict
        Includes all options (keys) and values from the section selected
        { key: value }
    """
    sec_dic = {}
    if sec in cfg:
        options = cfg[sec]
        for key, value in options.items():
            try:
                if convert == 'int':
                    sec_dic[key] = options.getint(key)
                elif convert == 'float':
                    sec_dic[key] = options.getfloat(key)
                elif convert == 'boolean':
                    sec_dic[key] = options.getboolean(key)
                else:
                    sec_dic[key] = value
            except:
                sec_dic[key] = value

    return sec_dic


def get_cfg_list(cfg, sec, opt, verbose=False):
    """
    Retrieve a comma separated list of values of a specified option.

    Parameters
    ----------
    cfg : configparser.ConfigParser()
        Configuration as retrieved by the function read_cfg_file().
    sec : str
        The section in which the option is located.
    opt : str
        The option that should be retrieved.
    verbose : bool, optional
        Print info, if either section or option could not be found in cfg.

    Returns
    -------
    list
        A list of values of the comma separated option.
    """
    list_ = get_cfg_option(cfg, sec, opt, verbose=verbose)
    if not list_:
        return []
    else:
        return [l.strip() for l in list_.split(',')]


def get_cfg_class(cfg, sec, mod_opt='module', cls_opt='class', std_mod=None,
                  std_cls=None, verbose=False):
    """
    Retrieve a class object as described by an option of a section.

    Parameters
    ----------
    cfg : configparser.ConfigParser()
        Configuration as retrieved by the function read_cfg_file().
    sec : str
        The section in which the option is located.
    mod_opt : str, optional
        The name of the option, which defines the module to load the class from
        (default: 'module').
    cls_opt : str, optional
        The name of the option, wich defines the class to load.
    std_mod : str, optional
        The module to load the class from. Is only used, if mod_opt cannot be
        found in cfg.
    std_cls : str, optional
        The name of the class, if class can not be found
    verbose : bool, optional
        Print info, if section, option or module could not be found.

    Returns
    -------
    class
        The requested class object.
    """
    module_ = get_cfg_option(cfg, sec, mod_opt, verbose=verbose) or std_mod
    if not module_:
        verbose and print("Could not determine module from cfg and no "
                          "std_module given.")
        return None
    try:
        module = importlib.import_module(module_, __package__)
    except:
        verbose and print("Could not load module '%s'" % module_)
        return None

    class_ = get_cfg_option(cfg, sec, cls_opt) or std_cls
    if not class_:
        verbose and print("Could not determine class from cfg.")
        return None

    if not hasattr(module, class_):
        verbose and print("Class '%s' is not in '%s'" % (class_, module))
        return None

    cls = getattr(module, class_)
    return cls



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """
"""
Provides physical functions and values.
"""
from . import ureg
from .utilities import str2u

from scipy import complex_
from scipy import exp
from scipy import float_
from scipy import inf
from scipy import isinf
from scipy import pi
from scipy import sqrt

import warnings


MATERIALS = {'PS': 1050,  # kg/m³
             'Polystyrene': 1050,
             'polystyrene': 1050,
             'ps': 1050,
             'Silica': 2000,
             'silica': 2000,
             'SiO': 2000,
             'sio': 2000,
             'Titania': 4230,
             'titania': 4230,
             'TiO': 4230,
             'tio': 4230}


class Material(object):
    """
    Material of the particle.
    """
    def __init__(self, name, density=None, density_unit='kg/m**3'):
        """
        Specify the name of the material of the particle.

        If the name is unknown you need to specify the density, as well.
        """
        self.name = name
        if name in MATERIALS:
            self._density = MATERIALS[name]
            self._density_unit = 'kg/m**3'
        else:
            if density:
                self._density = density
                self._density_unit = str2u(density_unit)
            else:
                raise Exception('Unknown Material "{}" and density is not'
                                'given.'.format(name))

    def get_density(self, unit=None):
        if unit is None or unit == self._density_unit:
            conv = 1.0
        else:
            conv = ureg(self._density_unit).to(unit).magnitude
        return self._density * conv


def viscosity_H2O(temp):
    """
    Calculates the viscosity (Pa*s) of water at temperature T in Kelvin.

    Warning
    -------
    Function is a fit to data of the viscosity of water at different
    temperatures [0, 100].

    Do not use it outside this range!

    References
    ----------
    CRC Handbook of Chemistry and Physics
        88th Edition
        CRC Press; 88 edition (October 1, 2007)
    """
    a = 3.4175e-4
    b = -0.01276
    c = 3.5857e-5

    if not(273.15 <= temp <= 373.15):
        warnings.warn('This function is only valid within a temperature'
                      'range between [273.15, 373.15] Kelvin!\n'
                      'T = {0:1.3f}'.format(temp))

    return a / (1 + b * temp + c * temp**2)


def dviscosity_H2O(temp):
    """
    Calculates the slope of the viscosity of water at a given temperature.

    See also
    --------
    viscosity_H2O
    """
    # the values to calculate vsicosity in Pa*s from T in K
    a = 3.4175e-4
    b = -0.01276
    c = 3.5857e-5

    if not(273.15 <= temp <= 373.15):
        warnings.warn('This function is only valid within a temperature'
                      'range between [273.15, 373.15] Kelvin!\n'
                      'T = {0:1.3f}'.format(temp))

    out = -((a * (b + 2 * c * temp)) / (1 + b * temp + c * temp**2)**2)
    return out


def density_H2O(temp):
    """
    Returns the density of water in kg/m³ at the specified temperature in
    Kelvin.

    Warning
    -------
    This function relies on a 3rd-order polynomial-fit to experimental
    data within the temperature region [0, 100] degrees Celsius. Within that
    range the fit reproduces the data within < 0.03 percent!

    Do not use it outside this range!

    References
    ----------
    CRC Handbook of Chemistry and Physics
        88th Edition
        CRC Press; 88 edition (October 1, 2007)
    """
    T = temp - 273.15  # convert to celcius scale

    if not(0 <= T <= 100):
        warnings.warn('This function is only valid within a tempearture'
                      'range between [273.15, 373,15] Kelvin!\n'
                      'T = {0:1.3f}'.format(T))

    a = 1.5690e-05
    b = -0.00591743
    c = 0.02020695
    d = 999.949648

    rho = a * T**3 + b * T**2 + c * T + d

    return rho


def kinematic_viscosity_H2O(temp):
    """
    Return the kinematic viscosity of water (m²/s) at the specified
    temperature in Kelvin.

    Warning
    -------
    Note that this function relies on fits to experimental
    data within the temperature region [0, 100] degrees Celsius. Within that
    range the fit reproduces the data within < 0.03 percent!

    Do not use it outside this range!

    References
    ----------
    CRC Handbook of Chemistry and Physics
        88th Edition
        CRC Press; 88 edition (October 1, 2007)
    """
    if not(0 <= temp-273.15 <= 100):
        warnings.warn('This function is only valid within a tempearture'
                      'range between [273.15, 373,15] Kelvin!')
    eta = viscosity_H2O(temp)
    rho = density_H2O(temp)

    return eta / rho


def drag(radius, temp, freq=0.0, height=inf, density=None, viscosity=None,
         lateral=True, verbose=False):
    """
    Calculate the frequency-dependent viscous drag of a sphere close to an
    infinit plane surface.

    The function uses an approximation to Faxen's solution for the drag near a
    plane surface at frequency f for a sphere of radius R. The drag depends on
    the height (surface to bead-center distance) of the sphere above the
    surface. Further, it depends on the kinematic viscosity of the medium, thus
    you can provide density and viscosity of the medium. If not provided, the
    kinematic viscosity of water at the given temperature is calculated.

    Arguments
    ---------
    radius : float
        Radius of the sphere in meter.
    temp : float
        Absolute temperature in kelvin.
    height : float
        Surface to bead-center distance in meter. If height = inf, the function
        returns stokes drag.
    density : float or None
        Density of the medium in kg/m³. If None, the density of water at the
        given temperature is used.
    viscosity : float
        Viscosity of the medium in Pa*s.  If None, the viscosity of water at
        the given temperature is used.
    later : bool
        !Not implemented yet!
        If true, the lateral drag is calculated. If false, the axial one
        is calculated.

    [Ref]: Tolić-Nørrelykke et al. Rev. Sci. Instrum. 77, 103101 (2006)

    Nota bene:
    In contrast to [Ref.], here f_nu is directly calculated with the given
    viscosity.
    """
    R = float_(radius)
    T = float_(temp)
    f = float_(freq)
    l = float_(height) if height >= radius else radius

    if verbose and height < radius:
        warnings.warn('Height is less than specified radius: '
                      '{0:1.3e} < {1:1.3e}\n'
                      'Height is set to radius as fallback.'
                      ''.format(height, radius))

    # viscous drag far away from a surface at constant speed
    eta = viscosity if viscosity else viscosity_H2O(T)
    d_0 = 6 * pi * eta * R

    if density is None or viscosity is None:
        nu = kinematic_viscosity_H2O(T)
        if verbose:
            print('Kinematic viscosity of water is used: '
                  '{0:1.4e} m²/s'.format(nu))
    else:
        nu = viscosity / density
        if verbose:
            print('Given viscosity and density is used '
                  '{0:1.4e} Pa*s / {1:1.4e} kg/m³ = {2:1.4e} m²/s'
                  ''.format(viscosity, density, nu))

    # f_nu - characteristic frequency
    f_nu = nu / (pi * R * R)

    # relative frequency
    f_ = f / f_nu

    # viscous drag far away from a surface at sinusodial movement
    # at frequency f
    # [Ref.] Equ. D4
    d_stokes = d_0 * (1 + complex_(1-1j) * sqrt(f_) - complex_(1j) * (2/9) * f_)

    if isinf(l):
        return d_stokes

    if lateral:
        # viscous drag at distance l from the surface at sinusodial movement f
        # [Ref.] Equ. D6
        delta = R / sqrt(f_)

        expon = 1 - exp((-(2 * l - R) / delta) * complex_(1-1j))

        # workaround 1 / sqrt(complex_(0)) = inf + nan j
        try:
            expon[f_ == 0.0] = complex_(0)
        except:
            if f_ == 0.0:
                # assume f_ is a scalar
                expon = complex_(0)

        denom = (1 - (9/16) * (R/l) * (1 - sqrt(f_)/3 * complex_(1-1j) +
                                       f_ * complex_(2j/9) - (4/3) * expon))

        d = d_stokes / denom
    else:
        # viscous drag at distance l from the surface at sinusodial movement f
        # [Ref.] Equ. D6
        delta = R / sqrt(f_)

        expon = 1 - exp((-(2 * l - R) / delta) * complex_(1-1j))

        # workaround 1 / sqrt(complex_(0)) = inf + nan j
        try:
            expon[f_ == 0.0] = complex_(0)
        except:
            if f_ == 0.0:
                # assume f_ is a scalar
                expon = complex_(0)

        denom = (1 - (9/16) * (R/l) * (1 - sqrt(f_)/3 * complex_(1-1j) +
                                       f_ * complex_(2j/9) - (4/3) * expon))

        d = d_stokes / denom

    return d


def faxen_factor(height, radius, lateral=True):
    """
    Return the relative change of the stokes drag according to Faxen's law.

    References
    ----------
    Schäffer, E., et al.
        Surface forces and drag coefficients of microspheres near a plane
        surface measured with optical tweezers.
        Langmuir 23, 3654–3665 (2007).
    """
    h = height
    r = radius

    if lateral:
        # Formular (5) in Schäffer et al.
        factor = 1 / (1 - 9 / 16 * r / h
                      + 1 / 8 * (r / h)**3
                      - 45 / 256 * (r / h)**4
                      - 1 / 16 * (r / h)**5
                      )
    else:
        # Formular (6) in  Schäffer et al.
        factor = 1 / (1 - 9 / 8 * r / h
                      + 1 / 2 * (r / h)**3
                      - 57 / 100 * (r / h)**4
                      + 1 / 5 * (r / h)**5
                      + 7 / 200 * (r / h)**11
                      - 1 / 25 * (r / h)**12
                      )
    return factor


def oseen_factor(height, radius, distance):
    """
    Factor that corrects stokes drag for two walls with given distance.

    Arguments
    ---------
    height : float
        height above the lover surface
    radius : float
        radius of the particle
    distance : float
        distance between the two walls

    References
    ----------
    Dufresne, E. R. et al.
        Brownian dynamics of a sphere between parallel walls.
        Europhys. Lett. 53, 264–270 (2001).
        Equation 5
    """
    o = (faxen_factor(height, radius) +
         faxen_factor(distance - height, radius) - 1)
    return o



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 15:29:19 2016

@author: Tobias Jachowski
"""
import numpy as np
import persistent

from .. import helpers as hp


class CalibrationSource(persistent.Persistent):
    """
    Offers attributes to be used by a Calibration.

    Attributes
    ----------
    corrfactor : float
        The correction factor of the size of the bead (i.e. the relative
        stokes drag coefficient far in the solution).
    dsurf : float
        The value of positionZ, where the center of the bead would be on the
        surface (m).
    beta : numpy.ndarray(n)
        The displacement sensitivity at the position, where the center of the
        bead would be on the surface (m/V).
    kappa : numpy.ndarray(n)
        The stiffness at the position, where the center of the bead would be on
        the surface (N/m).
    mbeta : numpy.ndarray(n)
        The slope of the displacement sensitivity in units of
        beadcenter-surface distance (i.e. focalshift corrected units of
        positionZ, m/V/m).
    mkappa : numpy.ndarray(n)
        The slope of the stiffness in units of beadcenter-surface distance
        (i.e. focalshift corrected units of positionZ).
    radiusspec : float
        The radius, as it was specified by the manufacturer (m).
    focalshift : float
        The focalshift of the Setup.
    """
    name = "Standard calibration source"

    def __init__(self, beta=None, kappa=None, radiusspec=0.0, corrfactor=1.0,
                 focalshift=1.0, mbeta=None, mkappa=None, dsurf=0.0, name=None,
                 **kwargs):
        """
        Parameters
        ----------
        beta : tuple or array (n)
            The displacement sensitivity at the position, where the center of
            the bead would be on the surface (m/V).
        kappa : tuple or array (n)
            The stiffness at the position, where the center of the bead would
            be on the surface (N/m).
        radiusspec : float
            The radius, as it was specified by the manufacturer (m).
        focalshift : float
            The focalshift of the Setup.
        mbeta : tuple or array (n), optional
            The slope of the displacement sensitivity in units of
            beadcenter-surface distance (i.e. focalshift corrected units of
            positionZ, m/V/m), (numpy.zeros(n), default).
        mkappa : tuple or array (n), optional
            The slope of the stiffness in units of beadcenter-surface distance
            (i.e. focalshift corrected units of positionZ, N/m/m),
            (numpy.zeros(n), default).
        corrfactor : float, optional
            The correction factor of the size of the bead (i.e. the relative
            stokes drag coefficient far in the solution), (1.0, default).
        dsurf : float, optional
            The value of positionZ (m), where the center of the bead would have
            been on the surface during the calibration measurement (0.0,
            default).
        """
        # Set beta and kappa to generice values. The calibration factors are
        # choosen in such a way that modules relying on a calibration are still
        # (somewhat) functional.
        if beta is None:
            print("Created generic values for beta. Make sure to provide a "
                  "proper calibration, if needed.")
            beta = [1.0, 1.0, 1.0]  # m/V
        if kappa is None:
            print("Created generic values for kappa. Make sure to provide a "
                  "proper calibration, if needed.")
            kappa = [1.0, 1.0, 1.0]  # N/m
        self.beta = np.array(hp.listify(beta))
        self.kappa = np.array(hp.listify(kappa))
        if mbeta is None:
            mbeta = [0.0 for i in range(self.beta.size)]
        self.mbeta = np.array(hp.listify(mbeta))
        if mkappa is None:
            mkappa = [0.0 for i in range(self.kappa.size)]
        self.mkappa = np.array(hp.listify(mkappa))
        self.dsurf = dsurf
        self.radiusspec = radiusspec
        self.corrfactor = corrfactor
        self.focalshift = focalshift
        self.name = name or ''
